import pytest

from qililab.settings.digital.digital_compilation_bus_settings import DigitalCompilationBusSettings
from qililab.typings.enums import Line


class TestDigitalCompilationBusSettings:
    def test_init_raises_error_if_weights_have_different_length(self):
        pass
