# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DataUnavailable Exception class"""


class DataUnavailable(Exception):
    """Exception raised when a method is requesting data that is not available.

    Args:
        message (str): Optional message to be displayed
    """

    def __init__(self, message: str | None):
        super().__init__("Requesting data not available" if message is None else message)
